/*
**   NELDER
**
**   purpose: multivariate minimization routine
**
**   format:  { xmin, fret } = nelder( &f, x) ;
**
**   input:   &f : function or proc returning a scalar
**            f should be defined as
**            { y } = f(theta) ;
**            with theta : (k x 1) vector or
**                         (k x n) matrix
**                  y    :  scalar or (n x 1) vector of
**                          corresponding function values
**
**             x : (k x 1) start value
**
**   output:  xmin : (k x 1) localized minimum
**            fret : scalar = f(xmin)
**
**   globals: _ftol : tolerance (relative difference in function values)
**            _maxit: maximum number of iterations
**            _show:  -1: dont show, n: show results every n iterations
**
**   remarks:
**             See W.H. Press (1986) et all,
**             Numerical Recipes: Ch.10 : the Downhill Simplex Method
**             Cambridge University Press
**
**             implementation:    G. Draisma  25-09-92
**                                Econometric Institute
**                                Erasmus University Rotterdam
**   updates:
**             2 feb 97: function evaluation in do loop
**
*/

declare  _ftol  ?= 1.0E-6 ;
declare  _maxit ?= 500 ;
declare  _show  ?= 25 ;

proc (2) = nelder( &f, x ) ;
   local alp, bet, gam,
      del, fp, dim, ndx,
      i, ilo, ihi, inhi,
      iter,
      p, pbar,
      xp, fxp,
      xxp, fxxp,
      f:proc ;

   p = x ;
   if cols(p) /= 1 ;
      errorlog "nelder: start vector no column vector" ;
      retp(p,0) ;
   endif ;

   alp  = 1.0 ;  /* mirror factor */
   bet  = 0.4 ;  /* shrink factor */
   gam  = 2.5 ;  /* extend-on-succes factor */
   iter  = 0 ;
   dim  = rows(p) ;
   p    = p~(p + (0.01*eye(dim)).*(p+sqrt(_ftol))) ;
   del  = 1 ;
   fp   = ones(dim+1,1) ;
   i    = 0 ;
   do while i < dim+1 ;
       i     = i + 1 ;
       fp[i] = f(p[.,i]) ;
   endo ;
   do while (del > _ftol) and (iter <= _maxit);
        if (((iter % _show)  == 0) and (_show >0));
           "iter (do-loop)" ; iter ;
           "p    " ;    p ;
           "fp   " ;   fp';
        endif ;
        iter  = iter + 1 ;
        ndx  = sortind(fp) ;
        ihi  = ndx[dim+1] ;
        inhi = ndx[dim] ;
        ilo  = ndx[1] ;
        del  = abs(fp[ihi] - fp[ilo])./(abs(fp[ilo]) + _ftol);
        pbar = meanc(p[.,ndx[1:dim]]') ;
        xp   = (1+alp)*pbar  - alp*p[.,ihi] ;
        fxp  = f(xp) ;
        if fxp < fp[inhi] ;
           if fxp <= fp[ilo] ;
              xxp  = (1+gam)*xp - gam*p[.,ihi] ;
              fxxp = f(xxp) ;
              if fxxp < fp[ilo] ;
              @  "extend" iter ;  @
                 p[.,ihi]   =  xxp ;
                 fp[ihi]    = fxxp ;
                 continue ;
              endif ;
           endif ;
              @  "reflect" iter;  @
              p[.,ihi]  = xp ;
              fp[ihi]   = fxp ;
        else ;
           xp = bet*p[.,ihi] +(1-bet)*pbar ;
           fxp = f(xp) ;
           if fxp < fp[inhi] ;
              @ "shrink 1 dimension" iter ; @
              p[.,ihi] = xp ;
              fp[ihi]  = fxp ;
           else ;
              @ "shrink simplex" iter ; @
              p =  (bet*p + (1-bet)*p[.,ilo]);
              i    = 0 ;
              do while i < dim+1 ;
                  i     = i + 1 ;
                  fp[i] = f(p[.,i]) ;
              endo ;
              @ fp   = f(p); @
           endif ;
        endif ;
    endo ;
    if iter > _maxit ;
       errorlog("nelder: exceeded _maxit") ;
    endif ;
    retp(p[.,ilo], fp[ilo]);
endp ;

